@@ -1,7 +1,7 @@
 /***************************************************************************
  * Base operating system software installation and configuration.
  *
- *   Copyright (C) 2022 by AK-47, along with transplanted code:
+ *   Copyright (C) 2022-2023 by AK-47, along with transplanted code:
  *    - Copyright (C) 2003-2010 by Warren Woodford
  *    - Heavily edited, with permision, by anticapitalista for antiX 2011-2014.
  *    - Heavily revised by dolphin oracle, adrian, and anticaptialista 2018.
@@ -23,48 +23,36 @@
 #include <sys/stat.h>
 #include <sys/statvfs.h>
 #include <QDebug>
+#include <QDir>
 #include <QFileInfo>
-#include <QCryptographicHash>
-#include <QDirIterator>
 #include <QMessageBox>
+#include "mprocess.h"
+#include "msettings.h"
+#include "partman.h"
 #include "base.h"
 
+#define BASE_BLOCK  (4*KB)
+
 Base::Base(MProcess &mproc, PartMan &pman, Ui::MeInstall &ui,
     const QSettings &appConf, const QCommandLineParser &appArgs)
-    : QObject(ui.boxMain), proc(mproc), gui(ui), partman(pman)
+    : proc(mproc), gui(ui), partman(pman)
 {
-    mediacheck = appArgs.isSet("media-check");
-    if (!mediacheck) nomediacheck = appArgs.isSet("no-media-check");
+    pretend = appArgs.isSet("pretend");
     populateMediaMounts = appConf.value("POPULATE_MEDIA_MOUNTPOINTS").toBool();
     nocopy = appArgs.isSet("nocopy");
     sync = appArgs.isSet("sync");
-}
+    bufferRoot = appConf.value("ROOT_BUFFER", 1024).toLongLong() * MB;
+    bufferHome = appConf.value("HOME_BUFFER", 1024).toLongLong() * MB;
 
-void Base::scanMedia()
-{
-    QSettings liveInfo("/live/config/initrd.out", QSettings::IniFormat);
-    const QString &sqpath = '/' + liveInfo.value("SQFILE_PATH", "antiX").toString();
-    const QString &sqtoram = liveInfo.value("TORAM_MP", "/live/to-ram").toString() + sqpath;
-    const QString &sqloc = liveInfo.value("SQFILE_DIR", "/live/boot-dev/antiX").toString();
-
-    // Check the installation media for errors (skip if not required).
-    bool checkmd5 = false;
-    if (!mediacheck) {
-        QFile file("/live/config/proc-cmdline");
-        if (file.open(QFile::ReadOnly | QFile::Text)) {
-            while (!file.atEnd()) {
-                if(file.readLine().trimmed() == "checkmd5") {
-                    checkmd5 = true;
-                    break;
-                }
-            }
+    QFile fileCLine("/live/config/proc-cmdline");
+    if (fileCLine.open(QFile::ReadOnly | QFile::Text)) {
+        QString clopts("Live boot options:");
+        while (!fileCLine.atEnd()) {
+            clopts.append(' ');
+            clopts.append(fileCLine.readLine().trimmed());
         }
-    }
-    if(checkmd5) proc.log("No media check (checkmd5)");
-    else if(nomediacheck) proc.log("No media check");
-    else {
-        const QString &sqfile = liveInfo.value("SQFILE_NAME", "linuxfs").toString();
-        checkMediaMD5(QFile::exists(sqtoram+'/'+sqfile) ? sqtoram : sqloc, sqfile);
+        proc.log(clopts);
+        fileCLine.close();
     }
 
     bootSource = "/live/aufs/boot";
@@ -72,110 +60,92 @@ void Base::scanMedia()
         << "/live/aufs/etc" << "/live/aufs/lib" << "/live/aufs/libx32" << "/live/aufs/lib64"
         << "/live/aufs/media" << "/live/aufs/mnt" << "/live/aufs/opt" << "/live/aufs/root"
         << "/live/aufs/sbin" << "/live/aufs/usr" << "/live/aufs/var" << "/live/aufs/home";
+    PartMan::VolumeSpec &vspecRoot = partman.volSpecs["/"];
+    PartMan::VolumeSpec &vspecBoot = partman.volSpecs["/boot"];
+    PartMan::VolumeSpec &vspecHome = partman.volSpecs["/home"];
 
-    // calculate required disk space
+    if (!pretend) proc.setExceptionMode(QT_TR_NOOP("Cannot access installation media."));
+
+    // Boot space required.
     proc.exec("du", {"-scb", bootSource}, nullptr, true);
-    partman.bootSpaceNeeded = proc.readOut(true).section('\n', -1).section('\t', 0, 0).toLongLong();
+    vspecBoot.image = proc.readOut(true).section('\n', -1).section('\t', 0, 0).toLongLong();
+    vspecBoot.minimum = vspecBoot.image * 3; // Include 1 backup and 1 new initrd image.
+    vspecBoot.preferred = vspecBoot.minimum;
+    if (vspecBoot.preferred < 1*GB) vspecBoot.preferred = 1*GB;
 
+    // Root space required.
+    QSettings liveInfo("/live/config/initrd.out", QSettings::IniFormat);
+    const QString &sqpath = '/' + liveInfo.value("SQFILE_PATH", "antiX").toString();
+    const QString &sqtoram = liveInfo.value("TORAM_MP", "/live/to-ram").toString() + sqpath;
+    const QString &sqloc = liveInfo.value("SQFILE_DIR", "/live/boot-dev/antiX").toString();
     bool floatOK = false;
     QString infile = sqtoram + "/linuxfs.info";
     if (!QFile::exists(infile)) infile = sqloc + "/linuxfs.info";
     if (QFile::exists(infile)) {
         const QSettings squashInfo(infile, QSettings::IniFormat);
-        partman.rootSpaceNeeded = 1024 * squashInfo.value("UncompressedSizeKB").toFloat(&floatOK);
+        vspecRoot.image = squashInfo.value("UncompressedSizeKB").toFloat(&floatOK) * KB;
     }
     if (!floatOK) {
         rootSources.prepend("-scb");
         proc.exec("du", rootSources, nullptr, true);
         rootSources.removeFirst();
-        partman.rootSpaceNeeded = proc.readOut(true).section('\n', -1).section('\t', 0, 0).toLongLong();
+        vspecRoot.image = proc.readOut(true).section('\n', -1).section('\t', 0, 0).toLongLong();
     }
-    qDebug() << "Image size:" << partman.rootSpaceNeeded << floatOK << infile;
+    qDebug() << "Basic image:" << vspecRoot.image << floatOK << infile;
+
     // Account for persistent root.
     if (QFileInfo::exists("/live/perist-root")) {
         proc.shell("df /live/persist-root --output=used --total |tail -n1", nullptr, true);
-        const long long rootfs_size = proc.readOut().toLongLong() * 1024;
+        const long long rootfs_size = proc.readOut().toLongLong() * KB;
         qDebug() << "rootfs size is " << rootfs_size;
         // probaby conservative, as rootfs will likely have some overlap with linuxfs.
-        partman.rootSpaceNeeded += rootfs_size;
+        vspecRoot.image += rootfs_size;
     }
+    // Account for file system formatting
+    struct statvfs svfs;
+    if (statvfs("/live/linux", &svfs) == 0) {
+        sourceInodes = svfs.f_files - svfs.f_ffree; // Will also be used for progress bar.
+    }
+    vspecRoot.image += (sourceInodes * BASE_BLOCK);
+    vspecRoot.minimum = vspecRoot.image + (vspecRoot.image / 20); // +5% for general FS shenanigans.
+    vspecRoot.preferred = vspecRoot.minimum + bufferRoot;
+    qDebug() << "Source inodes:" << sourceInodes << "Assumed block size:" << BASE_BLOCK;
 
-    qDebug() << "Minimum space:" << partman.bootSpaceNeeded << "(boot)," << partman.rootSpaceNeeded << "(root)";
+    vspecHome.minimum = 64*MB; // TODO: Set minimum home dynamically.
+    vspecHome.preferred = vspecHome.minimum + bufferHome;
+    qDebug() << "Minimum space:" << vspecBoot.image << "(boot)," << vspecRoot.image << "(root)";
+    proc.setExceptionMode(nullptr);
 }
 
-void Base::checkMediaMD5(const QString &path, const QString &sqfs)
+bool Base::saveHomeBasic() noexcept
 {
-    checking = true;
-    const QString osplash = gui.labelSplash->text();
-    const QString &nsplash = tr("Checking installation media.")
-        + "<br/><font size=2>%1% - " + tr("Press ESC to skip.") + "</font>";
-    gui.labelSplash->setText(nsplash.arg(0));
-    qint64 btotal = 0;
-    struct FileHash {
-        QString path;
-        QByteArray hash;
-    };
-    QList<FileHash> hashes;
-    static const char *failmsg = QT_TR_NOOP("The installation media is corrupt.");
-    // Obtain a list of MD5 hashes and their files.
-    QDirIterator it(path, {"*.md5"}, QDir::Files);
-    QStringList missing(sqfs);
-    if (!QFile::exists("/live/config/did-toram") || QFile::exists("/live/config/toram-all")) {
-        missing << "vmlinuz" << "initrd.gz";
+    proc.setExceptionMode(nullptr);
+    proc.log(__PRETTY_FUNCTION__, MProcess::Section);
+    QString homedir("/");
+    const DeviceItem *mntit = partman.mounts.value("/home");
+    if (!mntit || mntit->willFormat()) {
+        mntit = partman.mounts.value("/");
+        homedir = "/home";
     }
-    while (it.hasNext()) {
-        QFile file(it.next());
-        if (!file.open(QFile::ReadOnly | QFile::Text)) throw failmsg;
-        while (!file.atEnd()) {
-            QString line(file.readLine().trimmed());
-            const QString &fname = line.section(' ', 1, 1, QString::SectionSkipEmpty);
-            missing.removeOne(fname);
-            FileHash sfhash = {
-                .path = it.path() + '/' + fname,
-                .hash = QByteArray::fromHex(line.section(' ', 0, 0, QString::SectionSkipEmpty).toUtf8())
-            };
-            hashes.append(sfhash);
-            btotal += QFileInfo(sfhash.path).size();
-            qApp->processEvents();
-        }
-    }
-    if(!missing.isEmpty()) throw failmsg;
-    // Check the MD5 hash of each file.
-    const size_t bufsize = 65536;
-    std::unique_ptr<char[]> buf(new char[bufsize]);
-    qint64 bprog = 0;
-    for(const FileHash &fh : hashes) {
-        QListWidgetItem *logEntry = proc.log("Check MD5: " + fh.path);
-        QFile file(fh.path);
-        if (!file.open(QFile::ReadOnly)) throw failmsg;
-        QCryptographicHash hash(QCryptographicHash::Md5);
-        while (checking && !file.atEnd()) {
-            const qint64 rlen = file.read(buf.get(), bufsize);
-            if(rlen < 0) throw failmsg;
-            hash.addData(buf.get(), rlen);
-            bprog += rlen;
-            gui.labelSplash->setText(nsplash.arg((bprog * 100) / btotal));
-            qApp->processEvents();
-        }
-        if (!checking) break;
-        if (hash.result() != fh.hash) throw failmsg;
-        proc.log(logEntry);
-    }
-    gui.labelSplash->setText(osplash);
-    if (!checking) proc.log("Check halted");
-    checking = false;
-}
+    if (!mntit || mntit->willFormat()) return true;
+    const QString &homedev = mntit->mappedDevice();
 
-void Base::haltCheck(bool silent)
-{
-    if (!checking) return;
-    if (!silent) {
-        QMessageBox::StandardButton rc = QMessageBox::warning(gui.boxMain, QString(),
-            tr("Are you sure you want to skip checking the installation media?"),
-            QMessageBox::Yes | QMessageBox::No, QMessageBox::No);
-        if(rc != QMessageBox::Yes) return;
+    // Just in case the device or mount point is in use elsewhere.
+    proc.exec("/usr/bin/umount", {"-q", homedev});
+    proc.exec("/usr/bin/umount", {"-q", "/mnt/antiX"});
+
+    // Store a listing of /home to compare with the user name given later.
+    mkdir("/mnt/antiX", 0755);
+    QString opts = "ro";
+    if (mntit->type == DeviceItem::Subvolume) opts += ",subvol="+mntit->curLabel;
+    bool ok = proc.exec("/bin/mount", {"-o", opts, homedev, "/mnt/antiX"});
+    if (ok) {
+        QDir hd("/mnt/antiX" + homedir);
+        ok = hd.exists() && hd.isReadable();
+        homes = hd.entryList(QDir::Dirs);
+        proc.exec("/usr/bin/umount", {"-l", "/mnt/antiX"});
     }
-    checking = false;
+    return ok;
 }
 
 void Base::install()
@@ -184,7 +154,9 @@ void Base::install()
     if (proc.halted()) return;
     proc.advance(1, 2);
 
-    if (!partman.willFormat("/")) {
+    const DeviceItem *mntit = partman.mounts.value("/");
+    const bool skiphome = !(mntit && mntit->willFormat());
+    if (skiphome) {
         // if root was not formatted and not using --sync option then re-use it
         // remove all folders in root except for /home
         proc.status(tr("Deleting old system"));
@@ -204,7 +176,7 @@ void Base::install()
     mkdir("/mnt/antiX/sys", 0755);
     mkdir("/mnt/antiX/run", 0755);
 
-    copyLinux();
+    copyLinux(skiphome);
 
     proc.advance(1, 1);
     proc.status(tr("Fixing configuration"));
@@ -259,7 +231,7 @@ void Base::install()
     }
 }
 
-void Base::copyLinux()
+void Base::copyLinux(bool skiphome)
 {
     proc.log(__PRETTY_FUNCTION__, MProcess::Section);
     if (proc.halted()) return;
@@ -273,14 +245,9 @@ void Base::copyLinux()
     if (sync) {
         prog = "rsync";
         args << "--delete";
-        if (!partman.willFormat("/")) args << "--filter" << "protect home/*";
+        if (skiphome) args << "--filter" << "protect home/*";
     }
     args << bootSource << rootSources << "/mnt/antiX";
-    struct statvfs svfs;
-    fsfilcnt_t sourceInodes = 1;
-    if (statvfs("/live/linux", &svfs) == 0) {
-        sourceInodes = svfs.f_files - svfs.f_ffree;
-    }
     proc.advance(80, sourceInodes);
     proc.status(tr("Copying new system"));
     // Placed here so the progress bar moves to the right position before the next step.
@@ -291,8 +258,8 @@ void Base::copyLinux()
     QListWidgetItem *logEntry = proc.log(joined, MProcess::Exec);
 
     QEventLoop eloop;
-    connect(&proc, QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished), &eloop, &QEventLoop::quit);
-    connect(&proc, &QProcess::readyRead, &eloop, &QEventLoop::quit);
+    QObject::connect(&proc, QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished), &eloop, &QEventLoop::quit);
+    QObject::connect(&proc, &QProcess::readyRead, &eloop, &QEventLoop::quit);
     proc.start(prog, args);
     long ncopy = 0;
     while (proc.state() != QProcess::NotRunning) {
@@ -300,8 +267,8 @@ void Base::copyLinux()
         ncopy += proc.readAllStandardOutput().count('\n');
         proc.status(ncopy);
     }
-    disconnect(&proc, &QProcess::readyRead, nullptr, nullptr);
-    disconnect(&proc, QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished), nullptr, nullptr);
+    QObject::disconnect(&proc, &QProcess::readyRead, nullptr, nullptr);
+    QObject::disconnect(&proc, QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished), nullptr, nullptr);
 
     const QByteArray &StdErr = proc.readAllStandardError();
     if (!StdErr.isEmpty()) {
